#include <ESDOM/ESDOM_define.hpp>
#include <ESDOM/ESDOMException.hpp>
#include "DocumentInst.hpp"
#include "DocumentTypeInst.hpp"
#include "NodeInst.hpp"
#include "NodeListInst.hpp"
#include "ElementInst.hpp"
#include "CommentInst.hpp"
#include "TextInst.hpp"
#include "CDATASectionInst.hpp"
#include "AttrInst.hpp"
#include "ProcessingInstructionInst.hpp"
#include <ESDOM/Node.hpp>
#include <string.h>

using namespace ESDOM;

DocumentInst::DocumentInst()
 : NodeInst( _ESDOM_TEXT("#document"), Node::DOCUMENT_NODE, false, true )
{
	m_doctype = 0;
	m_documentElement = 0;
	m_standalone = -1;
	m_version = "1.0";
	m_valid = false;
}

DocumentInst::DocumentInst( const DocumentInst &original )
 : NodeInst( original )
{
	m_doctype = 0;
	m_documentElement = 0;
	m_standalone = original.m_standalone;
	m_version = original.m_version;
	m_valid = false;
}

DocumentInst::~DocumentInst()
{
}

NodeInst * DocumentInst::cloneNodeInst(bool deep) const
{
	DocumentInst *newNodeInst;

	try{
		newNodeInst = new DocumentInst( *this ); if(!newNodeInst)throw 0;
	}catch(...){
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}

	if( !deep )
		return newNodeInst;

	NodeListInst *newChildNodes = m_childNodes->clone(true);
	unsigned long index, length;
	for( index = 0, length = newChildNodes->getLength(); index < length; index++ )
		newChildNodes->item(index)->m_parentNode = (NodeInst*)newNodeInst;

	newNodeInst->m_childNodes->removeRef();
	newNodeInst->m_childNodes = newChildNodes;
	newNodeInst->m_childNodes->addRef();

	for( index = 0, length = newChildNodes->getLength(); index < length; index++ )
	{
		if( newChildNodes->item(index)->getNodeType() == Node::DOCUMENT_TYPE_NODE )
			newNodeInst->m_doctype = (DocumentTypeInst*)newChildNodes->item(index);
		else if( newChildNodes->item(index)->getNodeType() == Node::ELEMENT_NODE )
			newNodeInst->m_documentElement = (ElementInst*)newChildNodes->item(index);
	}
	newNodeInst->m_valid = m_valid;
	return newNodeInst;
}

DocumentTypeInst * DocumentInst::getDoctype() const
{
	return m_doctype;
}

ElementInst * DocumentInst::getDocumentElement() const
{
	return m_documentElement;
}

bool DocumentInst::getStandalone() const
{
	return ( m_standalone > 0 );
}

void DocumentInst::setStandalone(char n)
{
	if( n > 0 )
		m_standalone = 1;
	else if( n < 0 )
		m_standalone = -1;
	else
		m_standalone = 0;
}

ElementInst * DocumentInst::createElement(const ESDOM_CHAR *tagName) const
{
	// check >>
	if( !tagName || !ESDOMString::isName(tagName) )
		throw ESDOMException( ESDOMException::INVALID_CHARACTER_ERR );
	// <<

	ElementInst *newElement;
	try{
		newElement = new ElementInst(tagName); if(!newElement)throw 0;
	}catch(...){
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}

	return newElement;
}

TextInst * DocumentInst::createTextNode(const ESDOM_CHAR *data) const
{
	TextInst * newText;
	try{
		newText = new TextInst(); if(!newText)throw 0;
	}catch(...){
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}
	if( data )
		newText->setNodeValue( data );
	return newText;
}

CommentInst * DocumentInst::createComment(const ESDOM_CHAR *data) const
{
	CommentInst * newComment;
	try{
		newComment = new CommentInst(); if(!newComment)throw 0;
	}
	catch(...){
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}
	if( data )
		newComment->setNodeValue( data );
	return newComment;
}

CDATASectionInst * DocumentInst::createCDATASection(const ESDOM_CHAR *data) const
{
	CDATASectionInst * newCDATASection;
	try{
		newCDATASection = new CDATASectionInst; if(!newCDATASection)throw 0;
	}catch(...){
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}
	if( data )
		newCDATASection->setNodeValue( data );
	return newCDATASection;
}

ProcessingInstructionInst *	DocumentInst::createProcessingInstruction(const ESDOM_CHAR *target, const ESDOM_CHAR *data) const
{
	// check >>
	if( !target || !ESDOMString::isName(target) )
		throw ESDOMException( ESDOMException::INVALID_CHARACTER_ERR );
	// <<
	ProcessingInstructionInst * newPIInst;
	try{
		newPIInst = new ProcessingInstructionInst(target); if(!newPIInst)throw 0;
	}catch(...){
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}
	if( data )
		newPIInst->setNodeValue( data );
	return newPIInst;
}

AttrInst * DocumentInst::createAttribute(const ESDOM_CHAR *name) const
{
	// check >>
	if( !name || !ESDOMString::isName(name) )
		throw ESDOMException( ESDOMException::INVALID_CHARACTER_ERR );
	// <<

	AttrInst *newAttr;
	try{
		newAttr = new AttrInst(name); if(!newAttr)throw 0;
	}catch(...){
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}

	return newAttr;
}

NodeListInst * DocumentInst::getElementsByTagName(const ESDOM_CHAR *tagname) const
{
	if( m_documentElement
		&& tagname
		&& m_documentElement->getNodeName() != tagname
		&& memcmp( tagname ,_ESDOM_TEXT("*"), sizeof(ESDOM_CHAR)*2 ) )
		return m_documentElement->getElementsByTagName( tagname );

	NodeListInst *nodeList;
	try{
		nodeList = new NodeListInst(); if(!nodeList)throw 0;
	}catch(...){
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}

	if( !m_documentElement || !tagname )
		return nodeList;

	nodeList->appendNode( m_documentElement );

	NodeListInst *descendantNodeList = m_documentElement->getElementsByTagName( tagname );
	for( unsigned long index = 0, length = descendantNodeList->getLength(); index < length; index++ )
		nodeList->appendNode( descendantNodeList->item(index) );
	descendantNodeList->removeRef();
	return nodeList;
}

ElementInst * DocumentInst::getElementById(const ESDOM_CHAR *id) const
{
	if( !id || !m_documentElement )
		return 0;
	
	if( m_documentElement->getAttribute(_ESDOM_TEXT("id")) == id )
		return m_documentElement;
	return _getElementById( m_documentElement, id );
}

ESDOMString DocumentInst::setDoctype(const ESDOM_CHAR *doctype)
{
	if( !doctype || !*doctype )
	{
		if( m_doctype )
		{
			ESDOMString returnString = m_doctype->getNodeName();
			m_doctype->addRef();
			removeChild( m_doctype );
			m_doctype->removeRef();
			m_doctype = 0;
			return returnString;
		}
		return ESDOMString();
	}

	DocumentTypeInst *newDoctype;
	try{
		newDoctype = new DocumentTypeInst(doctype); if(!newDoctype)throw 0;
	}catch(...){
		throw ESDOMException( ESDOMException::OUT_OF_MEMORY );
	}

	if( m_doctype )
	{
		ESDOMString returnString = m_doctype->getNodeName();
		m_doctype->addRef();
		replaceChild(newDoctype, m_doctype);
		m_doctype->removeRef();
		m_doctype = newDoctype;
		return returnString;
	}

	m_doctype = newDoctype;
	if( m_documentElement )
		insertBefore(newDoctype, m_documentElement);
	else
		appendChild(newDoctype);

	return ESDOMString();
}

ESDOMString DocumentInst::getXmlVersion(void) const
{
	return m_version;
}

bool DocumentInst::isValid() const
{
	return m_valid;
}

////////////////////////////////////////////////

void DocumentInst::_setDocumentElement(ElementInst* element)
{
	if( !element )
		return;

	if( m_documentElement )
		replaceChild( element, m_documentElement );
	else
		appendChild( element );

	m_documentElement = element;
}

////////////////////////////////////////////////


ElementInst * DocumentInst::_getElementById(const ElementInst *element, const ESDOM_CHAR *id) const
{
	for( unsigned long index = 0, length = element->m_childNodes->getLength(); index < length; index++ )
	{
		if( element->m_childNodes->item(index)->getNodeType() == Node::ELEMENT_NODE )
		{
			if( ((ElementInst*)element->m_childNodes->item(index))->getAttribute(_ESDOM_TEXT("id")) == id )
			{
				return (ElementInst*)element->m_childNodes->item(index);
			}
			ElementInst *hitElement = _getElementById( (ElementInst*)element->m_childNodes->item(index), id );
			if( hitElement )
				return hitElement;
		}
	}
	return 0;
}
